@@ -1,3 +1,11 @@
+/*
+* Copyright (c) 2013-2020, The PurpleI2P Project
+*
+* This file is part of Purple i2pd project and licensed under BSD3
+*
+* See full license text in LICENSE file at top of project tree
+*/
+
 #ifndef CRYPTO_H__
 #define CRYPTO_H__
 
@@ -27,14 +35,14 @@
 #   define X509_getm_notAfter X509_get_notAfter
 #else
 #   define LEGACY_OPENSSL 0
-#	define OPENSSL_HKDF 1
 #   if (OPENSSL_VERSION_NUMBER >= 0x010101000) // 1.1.1
-#	   define OPENSSL_EDDSA 1
-#	   define OPENSSL_X25519 1
-#	   define OPENSSL_SIPHASH 1
+#       define OPENSSL_HKDF 1
+#       define OPENSSL_EDDSA 1
+#       define OPENSSL_X25519 1
+#       define OPENSSL_SIPHASH 1
 #   endif
 #   if !defined OPENSSL_NO_CHACHA && !defined OPENSSL_NO_POLY1305 // some builds might not include them
-#	   define OPENSSL_AEAD_CHACHA20_POLY1305 1 
+#       define OPENSSL_AEAD_CHACHA20_POLY1305 1
 #   endif
 #endif
 
@@ -80,20 +88,25 @@ namespace crypto
 			void GenerateKeys ();
 			const uint8_t * GetPublicKey () const { return m_PublicKey; };
 			void GetPrivateKey (uint8_t * priv) const;
-			void Agree (const uint8_t * pub, uint8_t * shared);			
+			void SetPrivateKey (const uint8_t * priv, bool calculatePublic = false);
+			void Agree (const uint8_t * pub, uint8_t * shared);
 
+			bool IsElligatorIneligible () const { return m_IsElligatorIneligible; }
+			void SetElligatorIneligible () { m_IsElligatorIneligible = true; }
+			
 		private:
 
-			uint8_t m_PublicKey[32];		
+			uint8_t m_PublicKey[32];
 #if OPENSSL_X25519
 			EVP_PKEY_CTX * m_Ctx;
 			EVP_PKEY * m_Pkey;
-#else			
+#else
 			BN_CTX * m_Ctx;
 			uint8_t m_PrivateKey[32];
-#endif			
+#endif
+			bool m_IsElligatorIneligible = false; // true if definitly ineligible
 	};
-	
+
 	// ElGamal
 	void ElGamalEncrypt (const uint8_t * key, const uint8_t * data, uint8_t * encrypted, BN_CTX * ctx, bool zeroPadding = false);
 	bool ElGamalDecrypt (const uint8_t * key, const uint8_t * encrypted, uint8_t * data, BN_CTX * ctx, bool zeroPadding = false);
@@ -116,15 +129,15 @@ namespace crypto
 		void operator^=(const ChipherBlock& other) // XOR
 		{
 			if (!(((size_t)buf | (size_t)other.buf) & 0x03)) // multiple of 4 ?
-			{	
+			{
 				for (int i = 0; i < 4; i++)
 					reinterpret_cast<uint32_t *>(buf)[i] ^= reinterpret_cast<const uint32_t *>(other.buf)[i];
-			}	
+			}
 			else
-			{	
+			{
 				for (int i = 0; i < 16; i++)
 					buf[i] ^= other.buf[i];
-			}					
+			}
 		}
 	};
 
@@ -156,9 +169,6 @@ namespace crypto
 
 
 #ifdef __AES__
-	#ifdef ARM64AES
-		void init_aesenc(void) __attribute__((constructor));
-	#endif
 	class ECBCryptoAESNI
 	{
 		public:
@@ -296,10 +306,20 @@ namespace crypto
 
 // HKDF
 
-	void HKDF (const uint8_t * salt, const uint8_t * key, size_t keyLen, const std::string& info, uint8_t * out); // salt - 32, out - 64, info <= 32 
+	void HKDF (const uint8_t * salt, const uint8_t * key, size_t keyLen, const std::string& info, uint8_t * out, size_t outLen = 64); // salt - 32, out - 32 or 64, info <= 32
+
+// Noise
+
+	struct NoiseSymmetricState
+	{
+		uint8_t m_H[32] /*h*/, m_CK[64] /*[ck, k]*/;
+
+		void MixHash (const uint8_t * buf, size_t len);
+		void MixKey (const uint8_t * sharedSecret);	
+	};
 
 // init and terminate
-	void InitCrypto (bool precomputation);
+	void InitCrypto (bool precomputation, bool aesni, bool avx, bool force);
 	void TerminateCrypto ();
 }
 }
