@@ -34,28 +34,54 @@ namespace datagram
 
 	void DatagramDestination::SendDatagramTo(const uint8_t * payload, size_t len, const i2p::data::IdentHash & identity, uint16_t fromPort, uint16_t toPort)
 	{
-		if (m_Owner->GetIdentity ()->GetSigningKeyType () == i2p::data::SIGNING_KEY_TYPE_DSA_SHA1)
-		{
-			uint8_t hash[32];
-			SHA256(payload, len, hash);
-			m_Owner->Sign (hash, 32, m_Signature.data ());
-		}
-		else
-			m_Owner->Sign (payload, len, m_Signature.data ());
-
 		auto session = ObtainSession(identity);
-		auto msg = CreateDataMessage ({{m_From.data (), m_From.size ()}, {m_Signature.data (), m_Signature.size ()}, {payload, len}},
-			fromPort, toPort, false, !session->IsRatchets ()); // datagram
-		session->SendMsg(msg);
+		SendDatagram (session, payload, len, fromPort, toPort);
+		FlushSendQueue (session);
 	}
 
 	void DatagramDestination::SendRawDatagramTo(const uint8_t * payload, size_t len, const i2p::data::IdentHash & identity, uint16_t fromPort, uint16_t toPort)
 	{
 		auto session = ObtainSession(identity);
-		auto msg = CreateDataMessage ({{payload, len}}, fromPort, toPort, true, !session->IsRatchets ()); // raw
-		session->SendMsg(msg);
+		SendRawDatagram (session, payload, len, fromPort, toPort);
+		FlushSendQueue (session);
 	}
 
+	std::shared_ptr<DatagramSession> DatagramDestination::GetSession(const i2p::data::IdentHash & ident)
+	{
+		return ObtainSession(ident);
+	}	
+		
+	void DatagramDestination::SendDatagram (std::shared_ptr<DatagramSession> session, const uint8_t * payload, size_t len, uint16_t fromPort, uint16_t toPort)
+	{
+		if (session)
+		{
+			if (m_Owner->GetIdentity ()->GetSigningKeyType () == i2p::data::SIGNING_KEY_TYPE_DSA_SHA1)
+			{
+				uint8_t hash[32];
+				SHA256(payload, len, hash);
+				m_Owner->Sign (hash, 32, m_Signature.data ());
+			}
+			else
+				m_Owner->Sign (payload, len, m_Signature.data ());
+
+			auto msg = CreateDataMessage ({{m_From.data (), m_From.size ()}, {m_Signature.data (), m_Signature.size ()}, {payload, len}},
+				fromPort, toPort, false, !session->IsRatchets ()); // datagram
+			session->SendMsg(msg);
+		}	
+	}	
+
+	void DatagramDestination::SendRawDatagram (std::shared_ptr<DatagramSession> session, const uint8_t * payload, size_t len, uint16_t fromPort, uint16_t toPort)
+	{
+		if (session)
+			session->SendMsg(CreateDataMessage ({{payload, len}}, fromPort, toPort, true, !session->IsRatchets ())); // raw
+	}
+		
+	void DatagramDestination::FlushSendQueue (std::shared_ptr<DatagramSession> session)
+	{
+		if (session)
+			session->FlushSendQueue ();
+	}	
+		
 	void DatagramDestination::HandleDatagram (uint16_t fromPort, uint16_t toPort,uint8_t * const &buf, size_t len)
 	{
 		i2p::data::IdentityEx identity;
@@ -126,7 +152,7 @@ namespace datagram
 		const std::vector<std::pair<const uint8_t *, size_t> >& payloads,
 		uint16_t fromPort, uint16_t toPort, bool isRaw, bool checksum)
 	{
-		auto msg = NewI2NPMessage ();
+		auto msg = m_I2NPMsgsPool.AcquireShared ();
 		uint8_t * buf = msg->GetPayload ();
 		buf += 4; // reserve for length
 		size_t size = m_Gzip ? m_Deflator.Deflate (payloads, buf, msg->maxLen - msg->len) :
@@ -196,7 +222,6 @@ namespace datagram
 		const i2p::data::IdentHash & remoteIdent) :
 		m_LocalDestination(localDestination),
 		m_RemoteIdent(remoteIdent),
-		m_SendQueueTimer(localDestination->GetService()),
 		m_RequestingLS(false)
 	{
 	}
@@ -204,21 +229,21 @@ namespace datagram
 	void DatagramSession::Start ()
 	{
 		m_LastUse = i2p::util::GetMillisecondsSinceEpoch ();
-		ScheduleFlushSendQueue();
 	}
 
 	void DatagramSession::Stop ()
 	{
-		m_SendQueueTimer.cancel ();
 	}
 
 	void DatagramSession::SendMsg(std::shared_ptr<I2NPMessage> msg)
 	{
 		// we used this session
 		m_LastUse = i2p::util::GetMillisecondsSinceEpoch();
-		// schedule send
-		auto self = shared_from_this();
-		m_LocalDestination->GetService().post(std::bind(&DatagramSession::HandleSend, self, msg));
+		if (msg || m_SendQueue.empty ())
+			m_SendQueue.push_back(msg);
+		// flush queue right away if full
+		if (!msg || m_SendQueue.size() >= DATAGRAM_SEND_QUEUE_MAX_SIZE) 
+			FlushSendQueue();
 	}
 
 	DatagramSession::Info DatagramSession::GetSessionInfo() const
@@ -256,94 +281,113 @@ namespace datagram
 
 	std::shared_ptr<i2p::garlic::GarlicRoutingPath> DatagramSession::GetSharedRoutingPath ()
 	{
-		if (!m_RoutingSession || !m_RoutingSession->GetOwner ()) 
+		if (!m_RemoteLeaseSet || m_RemoteLeaseSet->IsExpired ())
 		{
-			if(!m_RemoteLeaseSet) {
-				m_RemoteLeaseSet = m_LocalDestination->FindLeaseSet(m_RemoteIdent);
-			}
-			if(!m_RemoteLeaseSet) {
-				// no remote lease set
-				if(!m_RequestingLS) {
+			m_RemoteLeaseSet = m_LocalDestination->FindLeaseSet(m_RemoteIdent);
+			if (!m_RemoteLeaseSet)
+			{
+				if(!m_RequestingLS) 
+				{
 					m_RequestingLS = true;
 					m_LocalDestination->RequestDestination(m_RemoteIdent, std::bind(&DatagramSession::HandleLeaseSetUpdated, this, std::placeholders::_1));
 				}
 				return nullptr;
-			}
-			m_RoutingSession = m_LocalDestination->GetRoutingSession(m_RemoteLeaseSet, true);
-		}
+			}	
+		}	
+
+		if (!m_RoutingSession || !m_RoutingSession->GetOwner ()) 
+		{
+			bool found = false;
+			for (auto& it: m_PendingRoutingSessions)
+				if (it->GetOwner ()) // found established session
+				{
+					m_RoutingSession = it;
+					m_PendingRoutingSessions.clear ();
+					found = true;
+					break;
+				}		
+			if (!found)
+			{	
+				m_RoutingSession = m_LocalDestination->GetRoutingSession(m_RemoteLeaseSet, true);
+				if (!m_RoutingSession->GetOwner ())
+					m_PendingRoutingSessions.push_back (m_RoutingSession);
+			}	
+		}	
+		
 		auto path = m_RoutingSession->GetSharedRoutingPath();
-		if(path) {
-			if (m_CurrentOutboundTunnel && !m_CurrentOutboundTunnel->IsEstablished()) {
+		if (path && m_RoutingSession->IsRatchets () &&
+		    m_LastUse > m_RoutingSession->GetLastActivityTimestamp ()*1000 + DATAGRAM_SESSION_PATH_TIMEOUT)
+		{	
+			m_RoutingSession->SetSharedRoutingPath (nullptr);
+			path = nullptr;
+		}
+				
+		if (path) 
+		{
+			if (path->outboundTunnel && !path->outboundTunnel->IsEstablished ())
+			{	
 				// bad outbound tunnel, switch outbound tunnel
-				m_CurrentOutboundTunnel = m_LocalDestination->GetTunnelPool()->GetNextOutboundTunnel(m_CurrentOutboundTunnel);
-				path->outboundTunnel = m_CurrentOutboundTunnel;
-			}
-			if(m_CurrentRemoteLease && m_CurrentRemoteLease->ExpiresWithin(DATAGRAM_SESSION_LEASE_HANDOVER_WINDOW)) {
+				path->outboundTunnel = m_LocalDestination->GetTunnelPool()->GetNextOutboundTunnel(path->outboundTunnel);
+				if (!path->outboundTunnel)
+					m_RoutingSession->SetSharedRoutingPath (nullptr);
+			}	
+			
+			if (path->remoteLease && path->remoteLease->ExpiresWithin(DATAGRAM_SESSION_LEASE_HANDOVER_WINDOW)) 
+			{
 				// bad lease, switch to next one
-				if(m_RemoteLeaseSet && m_RemoteLeaseSet->IsExpired())
-					m_RemoteLeaseSet = m_LocalDestination->FindLeaseSet(m_RemoteIdent);
-				if(m_RemoteLeaseSet) {
-					auto ls = m_RemoteLeaseSet->GetNonExpiredLeasesExcluding([&](const i2p::data::Lease& l) -> bool {
-							return l.tunnelID == m_CurrentRemoteLease->tunnelID;
-					});
+				if (m_RemoteLeaseSet) 
+				{
+					auto ls = m_RemoteLeaseSet->GetNonExpiredLeasesExcluding(
+						[&](const i2p::data::Lease& l) -> bool 
+						{
+							return l.tunnelID == path->remoteLease->tunnelID;
+						});
 					auto sz = ls.size();
-					if (sz) {
+					if (sz) 
+					{
 						auto idx = rand() % sz;
-						m_CurrentRemoteLease = ls[idx];
+						path->remoteLease = ls[idx];
 					}
-				} else {
+					else
+						m_RoutingSession->SetSharedRoutingPath (nullptr);
+				} 
+				else 
+				{	
 					// no remote lease set?
 					LogPrint(eLogWarning, "DatagramSession: no cached remote lease set for ", m_RemoteIdent.ToBase32());
-				}
-				path->remoteLease = m_CurrentRemoteLease;
+					m_RoutingSession->SetSharedRoutingPath (nullptr);
+				}	
 			}
-		} else {
+		} 
+		else 
+		{
 			// no current path, make one
 			path = std::make_shared<i2p::garlic::GarlicRoutingPath>();
-			// switch outbound tunnel if bad
-			if(m_CurrentOutboundTunnel == nullptr || ! m_CurrentOutboundTunnel->IsEstablished()) {
-				m_CurrentOutboundTunnel = m_LocalDestination->GetTunnelPool()->GetNextOutboundTunnel(m_CurrentOutboundTunnel);
-			}
-			// switch lease if bad
-			if(m_CurrentRemoteLease && m_CurrentRemoteLease->ExpiresWithin(DATAGRAM_SESSION_LEASE_HANDOVER_WINDOW)) {
-				if(!m_RemoteLeaseSet) {
-					m_RemoteLeaseSet = m_LocalDestination->FindLeaseSet(m_RemoteIdent);
-				}
-				if(m_RemoteLeaseSet) {
-					// pick random next good lease
-					auto ls = m_RemoteLeaseSet->GetNonExpiredLeasesExcluding([&] (const i2p::data::Lease & l) -> bool {
-							if(m_CurrentRemoteLease)
-								return l.tunnelGateway == m_CurrentRemoteLease->tunnelGateway;
-							return false;
-					});
-					auto sz = ls.size();
-					if(sz) {
-						auto idx = rand() % sz;
-						m_CurrentRemoteLease = ls[idx];
-					}
-				} else {
-					// no remote lease set currently, bail
-					LogPrint(eLogWarning, "DatagramSession: no remote lease set found for ", m_RemoteIdent.ToBase32());
-					return nullptr;
-				}
-			} else if (!m_CurrentRemoteLease) {
-				if(!m_RemoteLeaseSet) m_RemoteLeaseSet = m_LocalDestination->FindLeaseSet(m_RemoteIdent);
-				if (m_RemoteLeaseSet)
+			path->outboundTunnel = m_LocalDestination->GetTunnelPool()->GetNextOutboundTunnel();
+			if (!path->outboundTunnel) return nullptr;
+				
+			if (m_RemoteLeaseSet) 
+			{
+				// pick random next good lease
+				auto ls = m_RemoteLeaseSet->GetNonExpiredLeases();
+				auto sz = ls.size();
+				if (sz) 
 				{
-					auto ls = m_RemoteLeaseSet->GetNonExpiredLeases();
-					auto sz = ls.size();
-					if (sz) {
-						auto idx = rand() % sz;
-						m_CurrentRemoteLease = ls[idx];
-					}
+					auto idx = rand() % sz;
+					path->remoteLease = ls[idx];
 				}
+				else
+					return nullptr;
+			} 
+			else 
+			{
+				// no remote lease set currently, bail
+				LogPrint(eLogWarning, "DatagramSession: no remote lease set found for ", m_RemoteIdent.ToBase32());
+				return nullptr;
 			}
-			path->outboundTunnel = m_CurrentOutboundTunnel;
-			path->remoteLease = m_CurrentRemoteLease;
 			m_RoutingSession->SetSharedRoutingPath(path);
 		}
 		return path;
-
 	}
 
 	void DatagramSession::HandleLeaseSetUpdated(std::shared_ptr<i2p::data::LeaseSet> ls)
@@ -356,16 +400,9 @@ namespace datagram
 		if(ls && ls->GetExpirationTime() > oldExpire) m_RemoteLeaseSet = ls;
 	}
 
-	void DatagramSession::HandleSend(std::shared_ptr<I2NPMessage> msg)
-	{
-		if (msg || m_SendQueue.empty ())
-			m_SendQueue.push_back(msg);
-		// flush queue right away if full
-		if(m_SendQueue.size() >= DATAGRAM_SEND_QUEUE_MAX_SIZE) FlushSendQueue();
-	}
-
 	void DatagramSession::FlushSendQueue ()
 	{
+		if (m_SendQueue.empty ()) return;
 		std::vector<i2p::tunnel::TunnelMessageBlock> send;
 		auto routingPath = GetSharedRoutingPath();
 		// if we don't have a routing path we will drop all queued messages
@@ -380,15 +417,6 @@ namespace datagram
 			routingPath->outboundTunnel->SendTunnelDataMsg(send);
 		}
 		m_SendQueue.clear();
-		ScheduleFlushSendQueue();
-	}
-
-	void DatagramSession::ScheduleFlushSendQueue()
-	{
-		boost::posix_time::milliseconds dlt(10);
-		m_SendQueueTimer.expires_from_now(dlt);
-		auto self = shared_from_this();
-		m_SendQueueTimer.async_wait([self](const boost::system::error_code & ec) { if(ec) return; self->FlushSendQueue(); });
 	}
 }
 }
