@@ -1,3 +1,11 @@
+/*
+* Copyright (c) 2013-2020, The PurpleI2P Project
+*
+* This file is part of Purple i2pd project and licensed under BSD3
+*
+* See full license text in LICENSE file at top of project tree
+*/
+
 #include <stdio.h>
 #include <string.h>
 #include "I2PEndian.h"
@@ -27,7 +35,7 @@ namespace data
 
 	RouterInfo::RouterInfo (const std::string& fullPath):
 		m_FullPath (fullPath), m_IsUpdated (false), m_IsUnreachable (false),
-		m_SupportedTransports (0), m_Caps (0)
+		m_SupportedTransports (0), m_Caps (0), m_Version (0)
 	{
 		m_Addresses = boost::make_shared<Addresses>(); // create empty list
 		m_Buffer = new uint8_t[MAX_RI_BUFFER_SIZE];
@@ -35,11 +43,12 @@ namespace data
 	}
 
 	RouterInfo::RouterInfo (const uint8_t * buf, int len):
-		m_IsUpdated (true), m_IsUnreachable (false), m_SupportedTransports (0), m_Caps (0)
+		m_IsUpdated (true), m_IsUnreachable (false), m_SupportedTransports (0),
+		m_Caps (0), m_Version (0)
 	{
 		m_Addresses = boost::make_shared<Addresses>(); // create empty list
 		if (len <= MAX_RI_BUFFER_SIZE)
-		{	
+		{
 			m_Buffer = new uint8_t[MAX_RI_BUFFER_SIZE];
 			memcpy (m_Buffer, buf, len);
 			m_BufferLen = len;
@@ -172,7 +181,6 @@ namespace data
 			LogPrint (eLogError, "RouterInfo: malformed message");
 			m_IsUnreachable = true;
 		}
-
 	}
 
 	void RouterInfo::ReadFromStream (std::istream& s)
@@ -191,7 +199,7 @@ namespace data
 			s.read ((char *)&address->cost, sizeof (address->cost));
 			s.read ((char *)&address->date, sizeof (address->date));
 			bool isNTCP2Only = false;
-			char transportStyle[6]; 
+			char transportStyle[6];
 			auto transportStyleLen = ReadString (transportStyle, 6, s) - 1;
 			if (!strncmp (transportStyle, "NTCP", 4)) // NTCP or NTCP2
 			{
@@ -230,13 +238,13 @@ namespace data
 						address->host.to_string (ecode);
 						if (!ecode)
 #endif
-						{	
+						{
 							// add supported protocol
 							if (address->host.is_v4 ())
 								supportedTransports |= (address->transportStyle == eTransportNTCP) ? eNTCPV4 : eSSUV4;
 							else
 								supportedTransports |= (address->transportStyle == eTransportNTCP) ? eNTCPV6 : eSSUV6;
-						}	
+						}
 					}
 				}
 				else if (!strcmp (key, "port"))
@@ -261,15 +269,15 @@ namespace data
 				{
 					if (!address->ntcp2) address->ntcp2.reset (new NTCP2Ext ());
 					supportedTransports |= (address->host.is_v4 ()) ? eNTCP2V4 : eNTCP2V6;
-					Base64ToByteStream (value, strlen (value), address->ntcp2->staticKey, 32);	
-				}	
+					Base64ToByteStream (value, strlen (value), address->ntcp2->staticKey, 32);
+				}
 				else if (!strcmp (key, "i")) // ntcp2 iv
 				{
 					if (!address->ntcp2) address->ntcp2.reset (new NTCP2Ext ());
 					supportedTransports |= (address->host.is_v4 ()) ? eNTCP2V4 : eNTCP2V6;
-					Base64ToByteStream (value, strlen (value), address->ntcp2->iv, 16);	
+					Base64ToByteStream (value, strlen (value), address->ntcp2->iv, 16);
 					address->ntcp2->isPublished = true; // presence if "i" means "published"
-				}	
+				}
 				else if (key[0] == 'i')
 				{
 					// introducers
@@ -277,7 +285,7 @@ namespace data
 					{
 						LogPrint (eLogError, "RouterInfo: Introducer is presented for non-SSU address. Skipped");
 						continue;
-					}	
+					}
 					introducers = true;
 					size_t l = strlen(key);
 					unsigned char index = key[l-1] - '0'; // TODO:
@@ -308,7 +316,7 @@ namespace data
 			}
 			if (introducers) supportedTransports |= eSSUV4; // in case if host is not presented
 			if (isNTCP2Only && address->ntcp2) address->ntcp2->isNTCP2Only = true;
-			if (supportedTransports) 
+			if (supportedTransports & ~(eNTCPV4 | eNTCPV6)) // exclude NTCP only
 			{
 				addresses->push_back(address);
 				m_SupportedTransports |= supportedTransports;
@@ -340,6 +348,21 @@ namespace data
 			// extract caps
 			if (!strcmp (key, "caps"))
 				ExtractCaps (value);
+			// extract version
+			else if (!strcmp (key, ROUTER_INFO_PROPERTY_VERSION))
+			{
+				m_Version = 0;
+				char * ch = value;
+				while (*ch)
+				{
+					if (*ch >= '0' && *ch <= '9')
+					{
+						m_Version *= 10;
+						m_Version += (*ch - '0');
+					}
+					ch++;
+				}
+			}
 			// check netId
 			else if (!strcmp (key, ROUTER_INFO_PROPERTY_NETID) && atoi (value) != i2p::context.GetNetID ())
 			{
@@ -368,9 +391,10 @@ namespace data
 			SetUnreachable (true);
 	}
 
-  bool RouterInfo::IsFamily(const std::string & fam) const {
-    return m_Family == fam;
-  }
+	bool RouterInfo::IsFamily(const std::string & fam) const
+	{
+		return m_Family == fam;
+	}
 
 	void RouterInfo::ExtractCaps (const char * value)
 	{
@@ -453,7 +477,12 @@ namespace data
 			s.write ((const char *)&address.date, sizeof (address.date));
 			std::stringstream properties;
 			if (address.transportStyle == eTransportNTCP)
-				WriteString (address.IsNTCP2 () ? "NTCP2" : "NTCP", s);
+			{
+				if (address.IsNTCP2 ())
+					WriteString ("NTCP2", s);
+				else
+					continue; // don't write NTCP address
+			}	
 			else if (address.transportStyle == eTransportSSU)
 			{
 				WriteString ("SSU", s);
@@ -564,7 +593,7 @@ namespace data
 				properties << '=';
 				WriteString (boost::lexical_cast<std::string>(address.port), properties);
 				properties << ';';
-			}	
+			}
 			if (address.IsNTCP2 ())
 			{
 				// publish s and v for NTCP2
@@ -572,7 +601,7 @@ namespace data
 				WriteString (address.ntcp2->staticKey.ToBase64 (), properties); properties << ';';
 				WriteString ("v", properties); properties << '=';
 				WriteString ("2", properties); properties << ';';
-			}	
+			}
 
 			uint16_t size = htobe16 (properties.str ().size ());
 			s.write ((char *)&size, sizeof (size));
@@ -680,20 +709,6 @@ namespace data
 		s.write (str.c_str (), len);
 	}
 
-	void RouterInfo::AddNTCPAddress (const char * host, int port)
-	{
-		auto addr = std::make_shared<Address>();
-		addr->host = boost::asio::ip::address::from_string (host);
-		addr->port = port;
-		addr->transportStyle = eTransportNTCP;
-		addr->cost = 6;
-		addr->date = 0;
-		for (const auto& it: *m_Addresses) // don't insert same address twice
-			if (*it == *addr) return;
-		m_SupportedTransports |= addr->host.is_v6 () ? eNTCPV6 : eNTCPV4;
-		m_Addresses->push_front(std::move(addr)); // always make NTCP first
-	}
-
 	void RouterInfo::AddSSUAddress (const char * host, int port, const uint8_t * key, int mtu)
 	{
 		auto addr = std::make_shared<Address>();
@@ -704,7 +719,10 @@ namespace data
 		addr->date = 0;
 		addr->ssu.reset (new SSUExt ());
 		addr->ssu->mtu = mtu;
-		memcpy (addr->ssu->key, key, 32);
+		if (key)
+			memcpy (addr->ssu->key, key, 32);
+		else
+			RAND_bytes (addr->ssu->key, 32);
 		for (const auto& it: *m_Addresses) // don't insert same address twice
 			if (*it == *addr) return;
 		m_SupportedTransports |= addr->host.is_v6 () ? eSSUV6 : eSSUV4;
@@ -726,7 +744,7 @@ namespace data
 		addr->ntcp2->isNTCP2Only = true; // NTCP2 only address
 		if (port) addr->ntcp2->isPublished = true;
 		memcpy (addr->ntcp2->staticKey, staticKey, 32);
-		memcpy (addr->ntcp2->iv, iv, 16);	
+		memcpy (addr->ntcp2->iv, iv, 16);
 		m_Addresses->push_back(std::move(addr));
 	}
 
@@ -793,14 +811,6 @@ namespace data
 		return "";
 	}
 
-	bool RouterInfo::IsNTCP (bool v4only) const
-	{
-		if (v4only)
-			return m_SupportedTransports & eNTCPV4;
-		else
-			return m_SupportedTransports & (eNTCPV4 | eNTCPV6);
-	}
-
 	bool RouterInfo::IsSSU (bool v4only) const
 	{
 		if (v4only)
@@ -838,7 +848,7 @@ namespace data
 			m_SupportedTransports |= eNTCPV6 | eSSUV6 | eNTCP2V6;
 	}
 
-  void RouterInfo::EnableV4 ()
+	void RouterInfo::EnableV4 ()
 	{
 		if (!IsV4 ())
 			m_SupportedTransports |= eNTCPV4 | eSSUV4 | eNTCP2V4;
@@ -861,7 +871,7 @@ namespace data
 		}
 	}
 
-  void RouterInfo::DisableV4 ()
+	void RouterInfo::DisableV4 ()
 	{
 		if (IsV4 ())
 		{
@@ -883,15 +893,6 @@ namespace data
 		return m_Caps & Caps::eUnreachable; // non-reachable
 	}
 
-	std::shared_ptr<const RouterInfo::Address> RouterInfo::GetNTCPAddress (bool v4only) const
-	{
-		return GetAddress (
-			[v4only](std::shared_ptr<const RouterInfo::Address> address)->bool
-			{
-				return (address->transportStyle == eTransportNTCP) && !address->IsNTCP2Only () && (!v4only || address->host.is_v4 ());
-			});
-	}
-
 	std::shared_ptr<const RouterInfo::Address> RouterInfo::GetSSUAddress (bool v4only) const
 	{
 		return GetAddress (
@@ -910,7 +911,7 @@ namespace data
 			});
 	}
 
-	template<typename Filter>	
+	template<typename Filter>
 	std::shared_ptr<const RouterInfo::Address> RouterInfo::GetAddress (Filter filter) const
 	{
 		// TODO: make it more generic using comparator
@@ -921,7 +922,7 @@ namespace data
 #endif
 		for (const auto& address : *addresses)
 			if (filter (address)) return address;
-		
+
 		return nullptr;
 	}
 
@@ -947,5 +948,12 @@ namespace data
 		if (encryptor)
 			encryptor->Encrypt (data, encrypted, ctx, true);
 	}
+
+	bool RouterInfo::IsEligibleFloodfill () const 
+	{
+		// floodfill must be reachable, >= 0.9.28 and not DSA
+		return IsReachable () && m_Version >= NETDB_MIN_FLOODFILL_VERSION &&
+			GetIdentity ()->GetSigningKeyType () != SIGNING_KEY_TYPE_DSA_SHA1; 
+	}	
 }
 }
