@@ -1,5 +1,14 @@
+/*
+* Copyright (c) 2013-2020, The PurpleI2P Project
+*
+* This file is part of Purple i2pd project and licensed under BSD3
+*
+* See full license text in LICENSE file at top of project tree
+*/
+
 #include <string.h>
 #include "I2PEndian.h"
+#include <random>
 #include <thread>
 #include <algorithm>
 #include <vector>
@@ -13,9 +22,6 @@
 #include "Config.h"
 #include "Tunnel.h"
 #include "TunnelPool.h"
-#ifdef WITH_EVENTS
-#include "Event.h"
-#endif
 
 namespace i2p
 {
@@ -34,9 +40,6 @@ namespace tunnel
 
 	void Tunnel::Build (uint32_t replyMsgID, std::shared_ptr<OutboundTunnel> outboundTunnel)
 	{
-#ifdef WITH_EVENTS
-		std::string peers = i2p::context.GetIdentity()->GetIdentHash().ToBase64();
-#endif
 		auto numHops = m_Config->GetNumHops ();
 		int numRecords = numHops <= STANDARD_NUM_RECORDS ? STANDARD_NUM_RECORDS : numHops;
 		auto msg = NewI2NPShortMessage ();
@@ -45,7 +48,7 @@ namespace tunnel
 		// shuffle records
 		std::vector<int> recordIndicies;
 		for (int i = 0; i < numRecords; i++) recordIndicies.push_back(i);
-		std::random_shuffle (recordIndicies.begin(), recordIndicies.end());
+		std::shuffle (recordIndicies.begin(), recordIndicies.end(), std::mt19937(std::random_device()()));
 
 		// create real records
 		uint8_t * records = msg->GetPayload () + 1;
@@ -63,15 +66,9 @@ namespace tunnel
 			hop->CreateBuildRequestRecord (records + idx*TUNNEL_BUILD_RECORD_SIZE, msgID, ctx);
 			hop->recordIndex = idx;
 			i++;
-#ifdef WITH_EVENTS
-			peers += ":" + hop->ident->GetIdentHash().ToBase64();
-#endif
 			hop = hop->next;
 		}
 		BN_CTX_free (ctx);
-#ifdef WITH_EVENTS
-		EmitTunnelEvent("tunnel.build", this, peers);
-#endif
 		// fill up fake records with random data
 		for (int i = numHops; i < numRecords; i++)
 		{
@@ -114,7 +111,7 @@ namespace tunnel
 		while (hop)
 		{
 			decryption.SetKey (hop->replyKey);
-			// decrypt records before and including current hop
+			// decrypt records before and current hop 
 			TunnelHopConfig * hop1 = hop;
 			while (hop1)
 			{
@@ -122,8 +119,22 @@ namespace tunnel
 				if (idx >= 0 && idx < msg[0])
 				{
 					uint8_t * record = msg + 1 + idx*TUNNEL_BUILD_RECORD_SIZE;
-					decryption.SetIV (hop->replyIV);
-					decryption.Decrypt(record, TUNNEL_BUILD_RECORD_SIZE, record);
+					if (hop1 == hop && hop1->IsECIES ())
+					{
+						uint8_t nonce[12];
+						memset (nonce, 0, 12);
+						if (!i2p::crypto::AEADChaCha20Poly1305 (record, TUNNEL_BUILD_RECORD_SIZE - 16, 
+							hop->m_H, 32, hop->m_CK, nonce, record, TUNNEL_BUILD_RECORD_SIZE - 16, false)) // decrypt
+						{
+							LogPrint (eLogWarning, "Tunnel: Response AEAD decryption failed");
+							return false;
+						}	
+					}
+					else
+					{	
+						decryption.SetIV (hop->replyIV);
+						decryption.Decrypt(record, TUNNEL_BUILD_RECORD_SIZE, record);
+					}	
 				}
 				else
 					LogPrint (eLogWarning, "Tunnel: hop index ", idx, " is out of range");
@@ -137,7 +148,7 @@ namespace tunnel
 		while (hop)
 		{
 			const uint8_t * record = msg + 1 + hop->recordIndex*TUNNEL_BUILD_RECORD_SIZE;
-			uint8_t ret = record[BUILD_RESPONSE_RECORD_RET_OFFSET];
+			uint8_t ret = record[hop->IsECIES () ? ECIES_BUILD_RESPONSE_RECORD_RET_OFFSET : BUILD_RESPONSE_RECORD_RET_OFFSET];
 			LogPrint (eLogDebug, "Tunnel: Build response ret code=", (int)ret);
 			auto profile = i2p::data::netdb.FindRouterProfile (hop->ident->GetIdentHash ());
 			if (profile)
@@ -206,9 +217,6 @@ namespace tunnel
 	void Tunnel::SetState(TunnelState state)
 	{
 		m_State = state;
-#ifdef WITH_EVENTS
-		EmitTunnelEvent("tunnel.state", this, state);
-#endif
 	}
 
 
@@ -309,6 +317,7 @@ namespace tunnel
 	{
 		for (auto& msg : msgs)
 		{
+			if (!msg.data) continue;
 			switch (msg.deliveryType)
 			{
 				case eDeliveryTypeLocal:
@@ -465,7 +474,7 @@ namespace tunnel
 	{
 		std::this_thread::sleep_for (std::chrono::seconds(1)); // wait for other parts are ready
 
-		uint64_t lastTs = 0;
+		uint64_t lastTs = 0, lastPoolsTs = 0;
 		while (m_IsRunning)
 		{
 			try
@@ -526,12 +535,20 @@ namespace tunnel
 					while (msg);
 				}
 
-				uint64_t ts = i2p::util::GetSecondsSinceEpoch ();
-				if (ts - lastTs >= 15) // manage tunnels every 15 seconds
+				if (i2p::transport::transports.IsOnline())
 				{
-					ManageTunnels ();
-					lastTs = ts;
-				}
+					uint64_t ts = i2p::util::GetSecondsSinceEpoch ();
+					if (ts - lastTs >= 15) // manage tunnels every 15 seconds
+					{
+						ManageTunnels ();
+						lastTs = ts;
+					}
+					if (ts - lastPoolsTs >= 5) // manage pools every 5 seconds
+					{
+						ManageTunnelPools (ts);
+						lastPoolsTs = ts;
+					}	
+				}	
 			}
 			catch (std::exception& ex)
 			{
@@ -573,7 +590,6 @@ namespace tunnel
 		ManageInboundTunnels ();
 		ManageOutboundTunnels ();
 		ManageTransitTunnels ();
-		ManageTunnelPools ();
 	}
 
 	void Tunnels::ManagePendingTunnels ()
@@ -590,7 +606,6 @@ namespace tunnel
 		for (auto it = pendingTunnels.begin (); it != pendingTunnels.end ();)
 		{
 			auto tunnel = it->second;
-			auto pool = tunnel->GetTunnelPool();
 			switch (tunnel->GetState ())
 			{
 				case eTunnelStatePending:
@@ -613,11 +628,6 @@ namespace tunnel
 								hop = hop->next;
 							}
 						}
-#ifdef WITH_EVENTS
-						EmitTunnelEvent("tunnel.state", tunnel.get(), eTunnelStateBuildFailed);
-#endif
-						// for i2lua
-						if(pool) pool->OnTunnelBuildResult(tunnel, eBuildResultTimeout);
 						// delete
 						it = pendingTunnels.erase (it);
 						m_NumFailedTunnelCreations++;
@@ -627,12 +637,6 @@ namespace tunnel
 				break;
 				case eTunnelStateBuildFailed:
 					LogPrint (eLogDebug, "Tunnel: pending build request ", it->first, " failed, deleted");
-#ifdef WITH_EVENTS
-					EmitTunnelEvent("tunnel.state", tunnel.get(), eTunnelStateBuildFailed);
-#endif
-					// for i2lua
-					if(pool) pool->OnTunnelBuildResult(tunnel, eBuildResultRejected);
-
 					it = pendingTunnels.erase (it);
 					m_NumFailedTunnelCreations++;
 				break;
@@ -797,16 +801,13 @@ namespace tunnel
 		}
 	}
 
-	void Tunnels::ManageTunnelPools ()
+	void Tunnels::ManageTunnelPools (uint64_t ts)
 	{
 		std::unique_lock<std::mutex> l(m_PoolsMutex);
 		for (auto& pool : m_Pools)
 		{
 			if (pool && pool->IsActive ())
-			{
-				pool->CreateTunnels ();
-				pool->TestTunnels ();
-			}
+				pool->ManageTunnels (ts);
 		}
 	}
 
